from distutils.version import LooseVersion

import airflow


AIRFLOW_VERSION_2 = LooseVersion(airflow.version.version) >= LooseVersion("2.0.0")
AIRFLOW_VERSION_1 = LooseVersion(airflow.version.version) < LooseVersion("2.0.0")

AIRFLOW_ABOVE_13 = LooseVersion(airflow.version.version) > LooseVersion("1.10.13")
AIRFLOW_ABOVE_10 = LooseVersion(airflow.version.version) > LooseVersion("1.10.10")
AIRFLOW_ABOVE_9 = LooseVersion(airflow.version.version) > LooseVersion("1.10.9")
AIRFLOW_ABOVE_6 = LooseVersion(airflow.version.version) > LooseVersion("1.10.6")

AIRFLOW_BELOW_10 = LooseVersion(airflow.version.version) < LooseVersion("1.10.10")



import logging
import os
import sys

from typing import Any, List, Optional, Union
from uuid import UUID

from airflow import DAG
from airflow.models import BaseOperator, DagModel

from dbnd import PythonTask, new_dbnd_context, override, parameter, task
from dbnd._core.configuration.dbnd_config import config
from dbnd._core.configuration.environ_config import (
    DBND_RESUBMIT_RUN,
    DBND_RUN_UID,
    ENV_DBND_DISABLE_SCHEDULED_DAGS_LOAD,
    in_quiet_mode,
)
from dbnd._core.configuration.scheduler_file_config_loader import (
    SchedulerFileConfigLoader,
)
from dbnd._core.constants import TaskExecutorType
from dbnd._core.run.databand_run import DatabandRun
from dbnd._core.settings import LoggingConfig, RunConfig
from dbnd._core.tracking.no_tracking import dont_track
from dbnd._core.tracking.schemas.tracking_info_run import ScheduledRunInfo
from dbnd._core.utils.basics.environ_utils import environ_enabled
from dbnd._core.utils.string_utils import clean_job_name
from dbnd._core.utils.timezone import convert_to_utc
from dbnd.api.scheduler import get_scheduled_jobs
from dbnd.tasks.basics.shell import bash_cmd


logger = logging.getLogger(__name__)


class DbndSchedulerDBDagsProvider(object):
    # by default only run the file sync if we are in the scheduler (and not the webserver)
    def __init__(self):
        self.scheduled_jobs = []

        if (
            config.getboolean("scheduler", "always_file_sync")
            or ("scheduler" in sys.argv)
        ) and not config.getboolean("scheduler", "never_file_sync"):
            self.file_config_loader = SchedulerFileConfigLoader()
            logger.debug("scheduler file syncing active")
        else:
            self.file_config_loader = None
            logger.debug("scheduler file syncing disabled")

        self.default_retries = config.getint("scheduler", "default_retries")

    def get_dags(self):  # type: () -> List[DAG]
        if not config.get("core", "databand_url"):
            self.scheduled_jobs = []
            return []
        logger.debug("about to get scheduler job dags from dbnd db")
        self.refresh_scheduled_jobs()
        dags = []
        for job in self.scheduled_jobs:
            if "schedule_interval" not in job:
                continue

            dag = self.job_to_dag(job)
            dag.sync_to_db()
            DagModel.get_dagmodel(dag.dag_id).set_is_paused(
                is_paused=not job["active"]
                or (
                    job.get("validation_errors", None) is not None
                    and len(job.get("validation_errors", None)) > 0
                ),
                including_subdags=False,
            )
            dags.append(dag)
        return dags

    def refresh_scheduled_jobs(self):
        changes = self.file_config_loader.sync() if self.file_config_loader else None
        if changes:
            logger.info(
                "[databand scheduler] changes in scheduler config file synced: %s"
                % ", ".join(("%s: %s" % (key, changes[key]) for key in changes))
            )

        self.scheduled_jobs = self.get_scheduled_jobs()

    def get_scheduled_jobs(self):  # type: () -> List[dict]
        return [
            s["DbndScheduledJob"]
            for s in get_scheduled_jobs()
            if not s["DbndScheduledJob"].get("validation_errors")
        ]

    def job_to_dag(self, job):  # type: (dict) -> Union[DAG, None]
        start_day = convert_to_utc(job.get("start_date", None))
        end_date = convert_to_utc(job.get("end_date", None))

        default_args = {
            "owner": job.get("create_user", None),
            "depends_on_past": job.get("depends_on_past", False),
            "start_date": job["start_date"],
            "end_date": end_date,
        }

        job_name = clean_job_name(job["name"])
        dag = DAG(
            "dbnd_launcher__%s" % job_name,
            start_date=start_day,
            default_args=default_args,
            schedule_interval=job.get("schedule_interval", None),
            catchup=job.get("catchup", False),
        )

        DbndSchedulerOperator(
            scheduled_cmd=job["cmd"],
            scheduled_job_name=job_name,
            with_name=False,
            scheduled_job_uid=job.get("uid", None),
            shell=config.getboolean("scheduler", "shell_cmd"),
            task_id="launcher",
            dag=dag,
            retries=job.get("retries", self.default_retries) or self.default_retries,
        )

        return dag


@dont_track
class DbndSchedulerOperator(BaseOperator):
    template_fields = ("scheduled_cmd",)
    template_ext = (".sh", ".bash")
    ui_color = "#f0ede4"

    def __init__(
        self,
        scheduled_cmd,
        scheduled_job_name,
        scheduled_job_uid,
        shell,
        with_name=True,
        **kwargs
    ):
        # type: (str, str, Optional[UUID], bool, bool, **Any) ->  DbndSchedulerOperator
        super(DbndSchedulerOperator, self).__init__(**kwargs)
        self.scheduled_job_name = scheduled_job_name
        self.scheduled_job_uid = scheduled_job_uid
        self.scheduled_cmd = scheduled_cmd
        self.shell = shell
        self.with_name = with_name

    def execute(self, context):
        scheduled_run_info = ScheduledRunInfo(
            scheduled_job_uid=self.scheduled_job_uid,
            scheduled_job_dag_run_id=context.get("dag_run").id,
            scheduled_date=context.get("task_instance").execution_date,
            scheduled_job_name=self.scheduled_job_name if self.with_name else None,
        )

        # disable heartbeat at this level,
        # otherwise scheduled jobs that will run on Kubernetes
        # will always have a heartbeat even if the actual driver
        # sent to Kubernetes is lost down the line,
        # which is the main purpose of the heartbeat
        with new_dbnd_context(
            name="airflow",
            conf={
                RunConfig.task_executor_type: override(TaskExecutorType.local),
                RunConfig.parallel: override(False),
                LoggingConfig.disabled: override(True),
            },
        ) as dc:
            launcher_task = Launcher(
                scheduled_cmd=self.scheduled_cmd,
                task_name=context.get("dag").dag_id,
                task_version="now",
                task_is_system=True,
                shell=self.shell,
            )

            dc.dbnd_run_task(
                task_or_task_name=launcher_task,
                scheduled_run_info=scheduled_run_info,
                send_heartbeat=False,
            )


class Launcher(PythonTask):
    scheduled_cmd = parameter[str]
    shell = parameter[bool]

    def run(self):
        env = os.environ.copy()
        env[DBND_RUN_UID] = str(DatabandRun.get_instance().run_uid)
        env[DBND_RESUBMIT_RUN] = "true"
        return bash_cmd.callable(
            cmd=self.scheduled_cmd, env=env, dbnd_env=False, shell=self.shell
        )


def get_dags():
    if environ_enabled(ENV_DBND_DISABLE_SCHEDULED_DAGS_LOAD):
        return None
    from dbnd._core.errors.base import DatabandConnectionException, DatabandApiError

    try:
        # let be sure that we are loaded
        config.load_system_configs()
        dags = DbndSchedulerDBDagsProvider().get_dags()

        if not in_quiet_mode():
            logger.info("providing %s dags from scheduled jobs" % len(dags))
        return dags
    except (DatabandConnectionException, DatabandApiError) as e:
        logger.error(str(e))
    except Exception as e:
        raise e



import logging

from dbnd._core.current import is_verbose, try_get_databand_run
from dbnd_airflow_contrib.utils.system_utils import (
    print_cpu_memory_usage,
    print_dmesg,
    print_stack_trace,
)


logger = logging.getLogger(__name__)


class KubernetesMetricsLogger(object):
    def __init__(self):
        pass

    def log_pod_running(self, task, node_name):
        task.log_system_metric("node_name", node_name)

    def log_pod_submitted(self, task, pod_name):
        import datetime

        timestamp = datetime.datetime.utcnow().isoformat()
        task.log_system_metric("pod_submitted_at", timestamp)
        task.log_system_metric("pod_name", pod_name)

    def log_pod_finished(self, task):
        import datetime

        timestamp = datetime.datetime.utcnow().isoformat()
        task.log_system_metric("pod_finished_at", timestamp)


def log_pod_events_on_sigterm(stack_frame):
    print_stack_trace(stack_frame)
    print_driver_events()
    print_cpu_memory_usage()
    if is_verbose():
        print_dmesg()


def print_driver_events():
    try:
        dbnd_run = try_get_databand_run()
        engine_config = dbnd_run.run_executor.remote_engine
        kube_client = engine_config.get_kube_client()
        from socket import gethostname

        driver_pod_name = gethostname()
        logger.info("Driver pod name is %s" % (driver_pod_name,))
        field_selector = "involvedObject.name=%s" % driver_pod_name
        logger.info("Field selector is %s" % (field_selector,))
        driver_events = kube_client.list_namespaced_event(
            namespace=engine_config.namespace, field_selector=field_selector
        )
        logger.info("Found %s driver events" % (len(driver_events.items),))
        for event in driver_events.items:
            message = create_log_message_from_event(event)
            logger.info(message)
    except Exception as e:
        logger.info("Could not retrieve driver events! Exception: %s", e)


def create_log_message_from_event(event):
    return """
    Pod: %s
    Event count: %s
    Message: %s
    Type: %s
    Action: %s
    Reason: %s
    Source: %s
    Event time: %s
    First timestamp: %s
    Last timestamp: %s
    Reporting component: %s
    Reporting instance: %s
    Involved object: %s
    """ % (
        event.metadata.name,
        event.count,
        event.message,
        event.type,
        event.action,
        event.reason,
        event.source,
        event.event_time,
        event.first_timestamp,
        event.last_timestamp,
        event.reporting_component,
        event.reporting_instance,
        event.involved_object,
    )



from dbnd_airflow.constants import AIRFLOW_ABOVE_10


def serialize_pod(pod, engine_config):
    """
    @param pod: Airflow pod definition
    @param engine_config: KubernetesEngineConfig
    @return:
    """
    if AIRFLOW_ABOVE_10:
        pod = pod.to_v1_kubernetes_pod()

        from airflow.kubernetes.kube_client import get_kube_client

        # airflow>1.10.10 uses official kubernetes client (https://github.com/kubernetes-client/python/)
        # to create pod json request instead of custom SimplePodRequest class
        kube_client = get_kube_client(in_cluster=engine_config.in_cluster)
        return kube_client.api_client.sanitize_for_serialization(pod)

    from airflow.contrib.kubernetes.kubernetes_request_factory.pod_request_factory import (
        SimplePodRequestFactory as AirflowSimplePodRequestFactory,
    )

    return AirflowSimplePodRequestFactory().create(pod)
